"""Basic functionalities for interacting with email servers."""

import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from os.path import basename
from typing import List, Literal, Optional, Union

from py4ai.core.logging import WithLogging


class EmailSender(WithLogging):
    """Email Sender Utility Class."""

    def __init__(
        self,
        email_address: str,
        username: str,
        password: str,
        smtp_address: str,
        auth_protocol: Optional[Literal["TLS", "SSL"]] = None,
        port: Optional[int] = None,
    ) -> None:
        """
        Return the instance of the class that acts as client to interact with an email server.

        :param email_address: Sender email address
        :param username: Username for authentication
        :param password: Password for authentication
        :param smtp_address: SMTP server address
        :param auth_protocol: Authentication protocol to use
        :param port: Port of SMTP server
        """
        self.email_address = email_address
        self.username = username
        self.password = password
        self.smtp_address = smtp_address
        self.auth_protocol = auth_protocol
        self.port = port

    def send_mail(
        self,
        text: str,
        subject: str,
        destination: str,
        attachments: Optional[List[str]] = None,
    ) -> None:
        """
        Send email.

        :param text: The text of the email
        :param subject: The subject of the email
        :param destination: The destination email address
        :param attachments: List with the files to send as attachments to the email
        :raises NotImplementedError: if given protocol is not implemented
        """
        msg = MIMEMultipart()
        msg["Subject"] = subject
        msg["From"] = self.email_address
        msg["To"] = destination
        msg.attach(MIMEText(text))
        if attachments is not None:
            for f in attachments:
                with open(f, "rb") as fil:
                    part = MIMEApplication(fil.read(), Name=basename(f))
                # After the file is closed
                part["Content-Disposition"] = 'attachment; filename="%s"' % basename(f)
                msg.attach(part)

        if self.auth_protocol == "SSL":
            port = 465 if self.port is None else self.port
            server: Union[smtplib.SMTP_SSL, smtplib.SMTP] = smtplib.SMTP_SSL(
                self.smtp_address, port=port
            )
        elif self.auth_protocol == "TLS":
            port = 587 if self.port is None else self.port
            server = smtplib.SMTP(self.smtp_address, port=port)
            server.starttls()
        elif self.auth_protocol is None:
            port = 25 if self.port is None else self.port
            server = smtplib.SMTP(self.smtp_address, port=port)
        else:
            raise NotImplementedError(f"{self.auth_protocol} not implemented")
        server.ehlo()
        server.login(self.username, self.password)
        server.sendmail(self.email_address, destination, msg.as_string())
        server.close()



"""Basic testing functionalities to be used in unittests."""

from time import time
from typing import Callable, Dict, TypeVar
from unittest import TestCase as CoreTestCase

import numpy as np
import pandas as pd

from py4ai.core.logging import WithLogging


class TestCase(CoreTestCase, WithLogging):
    """Tests class with basis functionalities."""

    def compareLists(self, first: list, second: list, strict: bool = False) -> None:
        """
        Compare lists.

        :param first: first list
        :param second: second list
        :param strict: whether to check element by element (strict) or only dimensions (non-strict)
        """
        if strict:
            for item1, item2 in zip(first, second):
                self.assertEqual(item1, item2)
        else:
            self.assertTrue(len(set(first).intersection(second)) == len(set(first)))

    def compareDicts(self, first: dict, second: dict, strict: bool = False) -> None:
        """
        Compare dictionaries.

        :param first: first dictionary
        :param second: second dictionary
        :param strict: whether to check element by element (strict) or only dimensions (non-strict)
        """
        for key in set(first.keys()).union(second.keys()):
            if isinstance(first[key], dict):
                self.compareDicts(first[key], second[key])
            elif isinstance(first[key], list):
                self.compareLists(first[key], second[key], strict)
            else:
                self.assertEqual(first[key], second[key])

    def compareDictsAlmostEqual(
        self, first: Dict[str, float], second: Dict[str, float], delta: float = 1e-12
    ):
        """
        Compare dictionaries and check if they are almost equal.

        :param first: first dictionary
        :param second: second dictionary
        :param delta: difference allowed between dictionaries' elements
        """
        self.assertEqual(len(first), len(second))
        self.assertListEqual(list(first.keys()), list(second.keys()))
        for first_value, second_value in zip(first.values(), second.values()):
            self.assertAlmostEqual(first_value, second_value, delta=delta)

    def compareDataFrames(
        self, first: pd.DataFrame, second: pd.DataFrame, msg: str
    ) -> None:
        """
        Compare dataframes.

        :param first: first dataframe
        :param second: second dataframe
        :param msg: message
        :raises failureException: if the dataframe are different
        """
        try:
            pd.testing.assert_frame_equal(first, second)
        except AssertionError as e:
            raise self.failureException("Input series are different") from e

    def compareSeries(self, first: pd.Series, second: pd.Series, msg: str) -> None:
        """
        Compare series.

        :param first: first series
        :param second: second series
        :param msg: message
        :raises failureException: if the dataframe are different
        """
        try:
            pd.testing.assert_series_equal(first, second)
        except AssertionError as e:
            raise self.failureException("Input series are different") from e

    def compareArrays(self, first: np.ndarray, second: np.ndarray, msg: str) -> None:
        """
        Compare arrays.

        :param first: first array
        :param second: second array
        :param msg: message
        :raises failureException: if the dataframe are different
        """
        try:
            np.testing.assert_almost_equal(first, second, decimal=7)
        except AssertionError as e:
            raise self.failureException("Input arrays are different") from e

    def setUp(self) -> None:
        """Set up the class to add custom comparing/assertion functionalities."""
        self.addTypeEqualityFunc(pd.DataFrame, self.compareDataFrames)
        self.addTypeEqualityFunc(pd.Series, self.compareSeries)
        self.addTypeEqualityFunc(np.ndarray, self.compareArrays)


T = TypeVar("T", bound=WithLogging)


def logTest(test: Callable[[T], None]) -> Callable[[T], None]:
    """
    Return a function to be used to decorate every test for adding logs on name of the tests and timings.

    :param test: test method to wrap
    :return: wrapped test method
    """

    def wrap(obj: T) -> None:
        t0 = time()
        obj.logger.info(f"Executing Test {str(test.__name__)}")
        test(obj)
        obj.logger.info(f"Execution Time: {time() - t0} secs")

    return wrap



"""Module with helpers to be used in async programming."""

import asyncio
from typing import Any, Coroutine, Optional, TypeVar

T = TypeVar("T")


class AsyncExecutor:
    """Helper class for computing async execution as sync."""

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """Instantiate the helper class.

        :param loop: EventLoop to be used to resolving co-routines. If None, the running EventLoop will be taken.
        """
        self.loop = asyncio.get_event_loop() if loop is None else loop

    def execute(self, f: Coroutine[Any, Any, T]) -> T:
        """Execute co-routing synchronously.

        :param f: co-routine to be executed.
        :returns: output of the co-routine.
        """
        t = self.loop.create_task(f)
        return self.loop.run_until_complete(t)



"""Module for providing interface to a cloud storage layer that persists files to be used in the application and downloaded on demand."""

import cgi
import os
from http.server import SimpleHTTPRequestHandler
from typing import Callable

import requests

from py4ai.core.logging import WithLogging
from py4ai.core.types import PathLike
from py4ai.core.utils.fs import create_dir_if_not_exists


class CloudSync(WithLogging):
    """Class for sync-ing to a remote file system."""

    def __init__(self, url: str, root: PathLike) -> None:
        """
        Download a file from the given url and saves it under the given root path.

        :param url: url to download file from
        :param root: local root path under which the downloaded file will be saved
        """
        self.url = url
        self.root = root

    def pathTo(self, filename: PathLike) -> PathLike:
        """
        Generate local path.

        :param filename: str
        :return: path to file
        """
        return os.path.join(self.root, filename)

    @staticmethod
    def create_base_directory(filename: PathLike) -> PathLike:
        """
        Create directory where the given file will be saved.

        :param filename: file name, str
        :return: path
        """
        return os.path.join(
            create_dir_if_not_exists(os.path.dirname(filename)),
            os.path.basename(filename),
        )

    def get(self, filename: PathLike) -> None:
        """
        Return file from the url.

        :param filename: name of file to be downloaded
        :raises FileNotFoundError: if file is not found
        """
        self.logger.info(f"Getting resource {filename} from {self.url}")

        r = requests.get(f"{self.url}/{filename}")

        if r.status_code == 200:
            file_out = self.pathTo(filename)

            with open(self.create_base_directory(file_out), "wb") as f:
                f.write(r.content)
        else:
            raise FileNotFoundError

    def get_if_not_exists(self, filename: PathLike) -> PathLike:
        """
        Get the file from the remote file system if not exists.

        :param filename: relative path
        :return: absolute path
        """
        file_out = self.pathTo(filename)

        if not os.path.exists(file_out):
            self.get(filename)
        return file_out

    def get_if_not_exists_decorator(
        self, f: Callable[[PathLike], PathLike]
    ) -> Callable[[PathLike], PathLike]:
        """
        Return a decorator to wrap functions that takes as inputs file to be downloaded from the remote source.

        :param f: function to be decorated, with signature f(filename)
        :return: wrapped function
        """

        def wrap(filename):
            return f(self.get_if_not_exists(filename))

        return wrap

    def upload(self, filename: PathLike) -> requests.Response:
        """
        Upload the file to the given url.

        :param filename: name of the file to be uploaded
        :return: `requests.Response`
        """
        with open(self.pathTo(filename), "rb") as fid:
            files = {"file": fid}
            self.logger.info(f"POST REQUEST ON {self.url}/{filename}")
            return requests.post(f"{self.url}/{filename}", files=files)


class HTTPRequestHandler(WithLogging, SimpleHTTPRequestHandler):
    """Performs POST operation."""

    def do_POST(self) -> None:
        """Post given resource."""
        path = self.translate_path(self.path)

        dirname = os.path.dirname(path)

        try:
            os.makedirs(dirname)
        except FileExistsError:
            pass

        form = cgi.FieldStorage(
            fp=self.rfile,
            headers=self.headers,  # type: ignore
            environ={
                "REQUEST_METHOD": "POST",
                "CONTENT_TYPE": self.headers["Content-Type"],
            },
        )

        filename = f"{dirname}/{form['file'].filename}" if path.endswith("/") else path
        data = form["file"].file.read()

        self.send_response(201, "Created")
        self.end_headers()

        with open(filename, "wb") as fid:
            fid.write(data)

        self.logger.info("POST Request handled")


if __name__ == "__main__":
    """
    python - m py4ai.utils.cloud --bind [IP_ADDRESS] [PORT]
    """

    import argparse
    from http.server import test  # type: ignore

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--bind",
        "-b",
        default="",
        metavar="ADDRESS",
        help="Specify alternate bind address " "[default: all interfaces]",
    )
    parser.add_argument(
        "port",
        action="store",
        default=8000,
        type=int,
        nargs="?",
        help="Specify alternate port [default: 8000]",
    )
    args = parser.parse_args()

    test(HandlerClass=HTTPRequestHandler, port=args.port, bind=args.bind)



"""Configuration classes."""
from datetime import datetime
from typing import Any, Hashable, List, Union

import pytz
from cfg_load import Configuration

from py4ai.core.types import PathLike
from py4ai.core.utils.dict import union


class BaseConfig(object):
    """Base configuration class."""

    def __init__(self, config: Configuration):
        """
        Class instance initializer.

        :param config: configuration
        """
        self.config = config

    def sublevel(self, name: Hashable) -> Configuration:
        """
        Return a sublevel of the main configuration.

        :param name: name of the sublevel
        :return: configuration of the sublevel
        """
        return Configuration(
            self.config[name], self.config.meta, self.config.meta["load_remote"]
        )

    def getValue(self, name: Hashable) -> Any:
        """
        Get the value of a configuration node.

        :param name: name of the configuration node
        :return: value of the configuratio node
        """
        return self.config[name]

    def safeGetValue(self, name: Hashable) -> Any:
        """
        Get the value of a configuration node, gracefully returning None if the node does not exist.

        :param name: name of the node
        :return: value of the node, or None if the node does not exist
        """
        return self.config.get(name, None)

    def update(self, other: Union[dict, Configuration]) -> "BaseConfig":
        """
        Update the current configuration.

        :param other: dictionary or Configuration containing the nodes of the configuration to be updated.
            In case other is a Configuration also metadata will be updated to other's metadata.
        :return: new configuration with the updated nodes
        :raises ValueError: if other is not a dict or a Configuration
        """
        if isinstance(other, dict):
            config = Configuration(
                union(self.config.to_dict(), other),
                union(
                    self.config.meta,
                    {
                        "updated_params": other,
                        "modification_datetime": datetime.now().astimezone(
                            tz=pytz.timezone("Europe/Rome")
                        ),
                    },
                ),
                self.config.meta["load_remote"],
            )
            return type(self)(config)
        elif isinstance(other, Configuration):
            newconfig = self.config.update(other)
            return type(self)(
                Configuration(
                    newconfig.to_dict(),
                    union(
                        newconfig.meta,
                        {
                            "updated_params": other.to_dict(),
                            "modification_datetime": datetime.now().astimezone(
                                tz=pytz.timezone("Europe/Rome")
                            ),
                        },
                    ),
                )
            )
        else:
            raise ValueError(f"Type {type(other)} cannot be merged to Configuration")


class FileSystemConfig(BaseConfig):
    """Configuration for file system paths."""

    @property
    def root(self) -> PathLike:
        """
        Return the root node value.

        :return: root node value
        """
        return self.getValue("root")

    def getFolder(self, path: Hashable) -> PathLike:
        """
        Return the folder name.

        :param path: name of the configuration node
        :return: folder name
        """
        return self.config["folders"][path]

    def getFile(self, file: Hashable) -> PathLike:
        """
        Return the file name.

        :param file: name of the configuration node
        :return: file name
        """
        return self.config["files"][file]


class AuthConfig(BaseConfig):
    """Authetication configuration."""

    @property
    def method(self) -> str:
        """
        Return the authentication method.

        :return: authentication method
        """
        return self.getValue("method")

    @property
    def filename(self) -> PathLike:
        """
        Return the name of the file containing the authentication details.

        :return: name of the file containing the authentication details
        """
        return self.getValue("filename")

    @property
    def user(self) -> str:
        """
        Return the user name.

        :return: user name
        """
        return self.getValue("user")

    @property
    def password(self) -> str:
        """
        Return the password.

        :return: password
        """
        return self.getValue("password")


class AuthService(BaseConfig):
    """Configuration for the authentication data."""

    @property
    def url(self) -> str:
        """
        Return the url of the authentication service.

        :return: url of the authentication service
        """
        return self.getValue("url")

    @property
    def check(self) -> str:
        """
        Return check.

        :return: check
        """
        return self.getValue("check")

    @property
    def decode(self) -> str:
        """
        Return decode.

        :return: decode
        """
        return self.getValue("decode")


class CheckService(BaseConfig):
    """Configuration for the check service."""

    @property
    def url(self) -> str:
        """
        Return the url of the check service.

        :return: url of the check service.
        """
        return self.getValue("url")

    @property
    def login(self) -> str:
        """
        Return the login url.

        :return: login url
        """
        return self.getValue("login")

    @property
    def logout(self) -> str:
        """
        Return the logout url.

        :return: logout url
        """
        return self.getValue("logout")


class AuthenticationServiceConfig(BaseConfig):
    """Configuration of the authentication service."""

    @property
    def secured(self) -> bool:
        """
        Return the secured flag.

        :return: secured flag
        """
        return self.getValue("secured")

    @property
    def ap_name(self) -> str:
        """
        Return the ap name.

        :return: ap name
        """
        return self.getValue("ap_name")

    @property
    def jwt_free_endpoints(self) -> List[str]:
        """
        Return the jwt free endpoints.

        :return: jwt free endpoints
        """
        return self.getValue("jwt_free_endpoints")

    @property
    def auth_service(self) -> AuthService:
        """
        Return the authentication data.

        :return: authentication data
        """
        return AuthService(self.sublevel("auth_service"))

    @property
    def check_service(self) -> CheckService:
        """
        Return the check service configuration.

        :return: check service configuration
        """
        return CheckService(self.sublevel("check_service"))

    @property
    def cors(self) -> str:
        """
        Return the cors.

        :return: cors
        """
        return self.getValue("cors")


class LoggingConfig(BaseConfig):
    """Logging configuration."""

    @property
    def level(self) -> str:
        """
        Returnn logging level.

        :return: level
        """
        return self.getValue("level")

    @property
    def filename(self) -> PathLike:
        """
        Name of the file where logs are stored.

        :return: filename
        """
        return self.getValue("filename")

    @property
    def default_config_file(self) -> PathLike:
        """
        Return default logging configuration file.

        :return: default config file
        """
        return self.getValue("default_config_file")

    @property
    def capture_warnings(self) -> bool:
        """
        Flag that determines whether waring are captured.

        :return: capture warnings
        """
        return self.getValue("capture_warnings")


class MongoConfig(BaseConfig):
    """Configuration for a Mongo DB."""

    @property
    def host(self) -> str:
        """
        Return ost name.

        :return: host name
        """
        return self.getValue("host")

    @property
    def port(self) -> int:
        """
        Return port.

        :return: port
        """
        return self.getValue("port")

    @property
    def db_name(self) -> str:
        """
        Return database name.

        :return: database name
        """
        return self.getValue("db_name")

    def getCollection(self, name) -> str:
        """
        Return collection name at a given configuration node.

        :param name: configuration node name
        :return: collection name
        """
        return self.config["collections"][name]

    @property
    def auth(self) -> AuthConfig:
        """
        Return authetication config.

        :return: authetication config
        """
        return AuthConfig(self.sublevel("auth"))

    @property
    def admin(self) -> AuthConfig:
        """
        Return administrator authentication config.

        :return: administrator authentication config
        """
        return AuthConfig(self.sublevel("admin"))

    @property
    def authSource(self) -> Any:
        """
        Return the authentication source.

        :return: authentication source
        """
        return self.safeGetValue("authSource")



"""Logging module."""
import os
import sys
from abc import ABC, abstractmethod
from importlib import import_module
from logging import FileHandler, Logger, captureWarnings, config, getLogger
from types import TracebackType
from typing import Any, Callable, List, Optional, Type, Union

from typing_extensions import Literal, TypedDict

from py4ai.core.config import merge_confs
from py4ai.core.types import PathLike

LevelTypes = Literal[
    "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET", 50, 40, 30, 20, 10, 0
]
StrLevelTypes = Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]


class LevelsDict(TypedDict):
    """Dictionary of logging levels."""

    CRITICAL: Literal[50]
    ERROR: Literal[40]
    WARNING: Literal[30]
    INFO: Literal[20]
    DEBUG: Literal[10]
    NOTSET: Literal[0]


DEFAULT_LOG_LEVEL: StrLevelTypes = "INFO"
levels = LevelsDict(CRITICAL=50, ERROR=40, WARNING=30, INFO=20, DEBUG=10, NOTSET=0)


class WithLoggingABC(ABC):
    """Abstract class providing logging capabilities."""

    @property
    @abstractmethod
    def logger(self) -> Logger:
        """Logger instance to be used to output logs within a class."""
        raise NotImplementedError


class WithLogging(WithLoggingABC):
    """Base class to be used for providing a logger embedded in the class."""

    @property
    def logger(self) -> Logger:
        """
        Create logger.

        :return: default logger
        """
        nameLogger = str(self.__class__).replace("<class '", "").replace("'>", "")
        return getLogger(nameLogger)

    def logResult(
        self, msg: Union[Callable[..., str], str], level: StrLevelTypes = "INFO"
    ) -> Callable[..., Any]:
        """
        Return a decorator to allow logging of inputs/outputs.

        :param msg: message to log
        :param level: logging level
        :return: wrapped method
        """

        def wrap(x: Any) -> Any:
            if isinstance(msg, str):
                self.logger.log(levels[level], msg)
            else:
                self.logger.log(levels[level], msg(x))
            return x

        return wrap


def getDefaultLogger(level: LevelTypes = levels[DEFAULT_LOG_LEVEL]) -> Logger:
    """
    Create default logger.

    :param level: logging level
    :return: root logger
    """
    logger = getLogger()
    logger.setLevel(level)
    return logger


def configFromFiles(
    config_files: List[PathLike],
    capture_warnings: bool = True,
    catch_exceptions: Optional[str] = None,
) -> None:
    """
    Configure loggers from configuration obtained merging configuration files.

    If any handler inherits from FileHandler create the directory for its output files if it does not exist yet.

    :param config_files: list of configuration files. Configurations in files with larger list index
        overwrite corresponding configurations in files with smaller list index.
    :param capture_warnings: whether to capture warnings with logger
    :param catch_exceptions: name of the logger used to catch exceptions. If None do not catch exception with loggers.
    """
    captureWarnings(capture_warnings)

    configuration = merge_confs(filenames=config_files, default=None)
    for v in configuration.to_dict()["handlers"].values():
        splitted = v["class"].split(".")
        if issubclass(
            getattr(import_module(".".join(splitted[:-1])), splitted[-1]), FileHandler
        ):
            if not os.path.exists(os.path.dirname(v["filename"])):
                os.makedirs(os.path.dirname(v["filename"]))
    config.dictConfig(configuration)

    if catch_exceptions is not None:
        except_logger = getLogger(catch_exceptions)
        print(
            f"Catching exceptions with {except_logger.name} logger using handlers "
            f'{", ".join([x.name for x in except_logger.handlers if x.name is not None])}'
        )

        def handle_exception(
            exc_type: Type[BaseException],
            exc_value: BaseException,
            exc_traceback: Optional[TracebackType],
        ) -> Any:
            if issubclass(exc_type, KeyboardInterrupt) and exc_traceback is not None:
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
            else:
                except_logger.error(
                    f"{exc_type.__name__}: {exc_value}",
                    exc_info=(exc_type, exc_value, exc_traceback),
                )

        sys.excepthook = handle_exception



"""Module that gathers useful decorators."""

import os
from functools import wraps
from glob import glob
from typing import Any, Callable, Iterable, Tuple, TypeVar

import pandas as pd

from py4ai.core.types import PathLike, T
from py4ai.core.utils.fs import create_dir_if_not_exists

A = TypeVar("A")
B = TypeVar("B")


def cache(func: Callable[..., T]) -> Callable[..., T]:
    """
    Return a decorator to cache function return values.

    :param func: input function
    :return: function wrapper
    """

    @wraps(func)
    def _wrap(obj):
        try:
            return obj.__dict__[func.__name__]
        except KeyError:
            score = func(obj)
            obj.__dict__[func.__name__] = score
            return score

    _wrap.__name__ = func.__name__
    return _wrap


def lazyproperty(obj: Any) -> property:
    """
    Return a lazy property, i.e. a property that is computed if needed and then cached.

    :param obj: method do decorate
    :return: wrapped method
    """
    return property(cache(obj))


class Cached(object):
    """Class to cache results and export them as pickle to be later reloaded."""

    @property
    def _cache(self):
        """
        Hidden property that stores the custom cache that can be exported and imported.

        :return: cached data
        """
        try:
            return self._cache_data
        except AttributeError:
            self._cache_data = {}
            return self._cache

    @staticmethod
    def cache(func: Callable[["Cached"], T]) -> property:
        """
        Return a decorator to cache function return values.

        :param func: input function
        :return: function wrapper
        """

        @wraps(func)
        def _wrap(obj: "Cached"):
            try:
                return obj._cache[func.__name__]
            except KeyError:
                score = func(obj)
                obj._cache[func.__name__] = score
                return score

        return property(_wrap)

    def clear_cache(self) -> None:
        """Clear cache of the object."""
        self._cache.clear()

    def save_pickles(self, path: PathLike) -> None:
        """
        Save pickle in given path.

        :param path: saving path
        """
        path = create_dir_if_not_exists(path)

        for k, data in self._cache.items():
            Cached.save_element(os.path.join(path, k), data)

    @staticmethod
    def save_element(filename: PathLike, obj: Any) -> None:
        """
        Save given object in given file.

        :param filename: saving path
        :param obj: object to be saved
        :raises Exception: if any error occurs while saving the file
        """
        if isinstance(obj, dict):
            create_dir_if_not_exists(filename)
            [Cached.save_element(os.path.join(filename, k), v) for k, v in obj.items()]
        elif isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
            pd.to_pickle(obj, "%s.p" % filename)
        else:
            try:
                pd.to_pickle(obj, "%s.p" % filename)
            except Exception as ex:
                raise type(ex)("Cannot save input of type %s" % str(obj.__class__))

    def load(self, filename: PathLike) -> None:
        """
        Load pickle at given path (or all pickles in given folder).

        :param filename: path to pickles
        """
        self._cache.update(dict(self.load_element(filename, "")))

    @classmethod
    def load_element(
        cls, filename: PathLike, prefix: str = ""
    ) -> Iterable[Tuple[PathLike, Any]]:
        """
        Load given element/property of the object.

        :param filename: name of the file/directory where the data should be taken from
        :param prefix: prefix to add to elements in filename in case it is a directory
        :yield: iterable of couples (path, loaded object)
        """
        if os.path.isdir(filename):
            for path in glob(os.path.join(filename, "*")):
                for name, obj in cls.load_element(
                    path, f"{cls._reformat_name(filename)}_"
                ):
                    yield f"{prefix}{name}", obj
        else:
            yield str(cls._reformat_name(filename)), pd.read_pickle(filename)

    @staticmethod
    def _reformat_name(filename: PathLike) -> PathLike:
        """
        Reformat file name to make it readable with load_element.

        :param filename: file name to reformat
        :return: reformatted filename
        """
        if os.path.isfile(filename):
            return os.path.basename(filename).replace(".p", "")
        else:
            return os.path.basename(filename)


def same_type(f: Callable[[A, B], T]) -> Callable[[A, B], T]:
    """
    Check that both arguments of input function have the same type.

    :param f: function
    :return: function
    """

    @wraps(f)
    def new_f(self, other):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"other's type ({type(other)}) is different from self's type ({type(self)})"
            )
        return f(self, other)

    return new_f
